#include <stdio.h>
#include <kallisti/libk.h>
#include <kallisti/abi/fs.h>
#include <kallisti/abi/ta.h>
#include <kallisti/abi/thread.h>

#include "decore.h"

#include "avi_read.h"

/* Library handles */
abi_fs_t    *fs = NULL;
abi_thread_t    *thread = NULL;
abi_ta_t *ta = NULL;

unsigned int tex;

#define MY_APP_ID 0x12345

DEC_PARAM dec_param;
DEC_SET dec_set;
DEC_FRAME dec_frame;

avi_t * my_avi;

/* allocate texture ram */

void init_tex()
{
    tex = ta->txr_allocate((1024*1024*3)/2);
}

/* display one frame on video, scaled up/down to 640 pixels wide */

void draw_frame(int x, int y)
{
    poly_hdr_t poly;
    vertex_ot_t vert;
    float scale;
    float yscaled;
    float ys;

    scale = 640.0f / x;

    yscaled = y * scale;

    if (yscaled < 480.0f)
	ys = (480.0f - yscaled)/2.0f;
    else
	ys = 1.0f;

    ta->poly_hdr_txr(&poly, TA_TRANSLUCENT, TA_YUV422, 1024, 1024, tex, TA_NO_FILTER);

    vert.r = vert.g = vert.b = vert.a = 1.0f;
    vert.oa = vert.or = vert.ob = 0.0f;
    vert.flags = TA_VERTEX_NORMAL;

    ta->begin_render();

    ta->commit_eol();

    ta->commit_poly_hdr(&poly);

    vert.x = 1.0f;
    vert.y = ys;
    vert.z = 1.0f;
    vert.u = 0.0f;
    vert.v = 0.0f;
    ta->commit_vertex(&vert, sizeof(vert));

    vert.x = 640.0f;
    vert.y = ys;
    vert.z = 1.0f;
    vert.u = x/1024.0f;
    vert.v = 0.0f;
    ta->commit_vertex(&vert, sizeof(vert));

    vert.x = 1.0f;
    vert.y = ys+yscaled-1.0f;
    vert.z = 1.0f;
    vert.u = 0.0f;
    vert.v = y/1024.0f;
    ta->commit_vertex(&vert, sizeof(vert));

    vert.flags = TA_VERTEX_EOL;
    vert.x = 640.0f;
    vert.y = ys+yscaled-1.0f;
    vert.z = 1.0f;
    vert.u = x/1024.0f;
    vert.v = y/1024.0f;
    ta->commit_vertex(&vert, sizeof(vert));

    ta->commit_eol();
    ta->finish_frame();
}

/* copy YUV411 data from libdivxdecore into a texture and convert to YUV422 */

void frame_copy(int w, int h, char *outbuf[3])
{
    int i,j;
    unsigned short *texp;
    unsigned char *y = outbuf[0];
    unsigned char *cb = outbuf[2];
    unsigned char *cr = outbuf[1];

    texp = ta->txr_map(tex);

    for(j = 0; j < h; j+=2) {
	for(i = 0; i < w; i+=2) {
	    texp[i] = *(cb++) | (((unsigned short)*(y++))<<8);
	    texp[i+1] = *(cr++) | (((unsigned short)*(y++))<<8);
	}
	texp += 1024;
	cb -= w/2;
	cr -= w/2;
	for(i = 0; i < w; i+=2) {
	    texp[i] = *(cb++) | (((unsigned short)*(y++))<<8);
	    texp[i+1] = *(cr++) | (((unsigned short)*(y++))<<8);
	}
	texp += 1024;
    }
}

int main()
{
    unsigned int start, end;
    int fd;
    char *vidbuf;
    char *outbuf[3];
    int rv;

    printf("Program starting\r\n");
    
    fs = lib_open("fs");
    if (fs == NULL) { printf("Can't open fs lib\r\n"); return -1; }
    ta = lib_open("ta");
    if (ta == NULL) { printf("Can't open ta lib\r\n"); return -1; }
    thread = lib_open("thread");
    if (thread == NULL) { printf("Can't open thread lib\r\n"); return -1; }

    fd = fs->open("/rd/test.avi", O_RDONLY);

    if (!AVI_is_avi(fd)) {
	printf("is not AVI\n");
    } else {
	printf("is AVI\n");
	
	my_avi = AVI_open_input(fd, 1);
	printf("%d x %d\n",AVI_video_width(my_avi),AVI_video_height(my_avi));
	printf("%d frames at %d fps\n",AVI_video_frames(my_avi),(int) AVI_frame_rate(my_avi));
	printf("compressed with %s\n",AVI_video_compressor(my_avi));

	/* really should check that compressor is "divx" here */
	
	dec_param.x_dim = AVI_video_width(my_avi);
	dec_param.y_dim = AVI_video_height(my_avi);
	dec_param.color_depth = 32;
	
	decore(MY_APP_ID, DEC_OPT_INIT, &dec_param, NULL);
	
	dec_set.postproc_level = 0;
	decore(MY_APP_ID, DEC_OPT_SETPP, &dec_set, NULL);
	
	vidbuf = malloc(512*1024);
	
	outbuf[0] = malloc(dec_param.x_dim*dec_param.y_dim);
	outbuf[1] = malloc((dec_param.x_dim*dec_param.y_dim)/4);
	outbuf[2] = malloc((dec_param.x_dim*dec_param.y_dim)/4);
	
	init_tex();
	
	start = thread->jiffies();
	
	rv = AVI_read_frame(my_avi, vidbuf);
	
	while(rv > 0) {
	    dec_frame.length = rv;
	    dec_frame.bitstream = vidbuf;
	    dec_frame.bmp = outbuf;
	    dec_frame.render_flag = 1;
	    
	    decore(MY_APP_ID, 0, &dec_frame, NULL);
	    
	    frame_copy(dec_param.x_dim, dec_param.y_dim, outbuf);

	    draw_frame(dec_param.x_dim, dec_param.y_dim);

	    rv = AVI_read_frame(my_avi, vidbuf);
	}
	end = thread->jiffies();
	printf("playback took %d jiffies\r\n",end-start);

	printf("average framerate was %d fps\r\n", (100*AVI_video_frames(my_avi))/(end-start));
	
	free(vidbuf);
	free(outbuf[0]);
	free(outbuf[1]);
	free(outbuf[2]);
    }

    printf("Done, returning\r\n");

    return 0;
}
